/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.glprog.glprog;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import com.bitctrl.Constants;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.iav.gllib.gllib.dav.GanglinieMQ;
import de.bsvrz.iav.gllib.gllib.dav.GanglinienMQOperationen;
import de.bsvrz.iav.gllib.gllib.dav.GlProgAnfrage;
import de.bsvrz.iav.gllib.gllib.modell.onlinedaten.OdPrognoseGanglinienAnfrage;
import de.bsvrz.iav.gllib.gllib.modell.onlinedaten.OdPrognoseGanglinienAntwort;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.bitctrl.modell.AnmeldeException;
import de.bsvrz.sys.funclib.bitctrl.modell.Datensatz.Status;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensendeException;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.kalender.objekte.EreignisListener;
import de.bsvrz.sys.funclib.bitctrl.modell.kalender.objekte.EreignisseAktualisiertEvent;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Ein Worker, der Ganglinienprognoseanfragen bearbeitet.
 * <p>
 * Für zyklische Prognossn, muss die Prognose geprüft werden, wenn sich Ereignisse ändern. Da
 * Ereignisse nicht pro Messquerschnitt überwacht werden können, wird die Menge aller Ereignisse auf
 * Änderung überwacht und ggf. alle zyklischen Prognosen überprüft.
 *
 * @author BitCtrl Systems GmbH, Schumann
 */
public class AnfrageInBearbeitung implements Runnable {

	private final EreignisListener ereignisListener = new EreignisListener() {

		/**
		 * Setzt das Flag <em>neueEreignisse</em> auf <code>true</code>.
		 */
		@Override
		public void ereignisseAktualisiert(final EreignisseAktualisiertEvent e) {
			setNeueEreignisse(true);
		}

	};

	/** Der Logger der Klasse. */
	private final Debug log = Debug.getLogger();

	/** Die zu bearbeitende Anfrage. */
	private final OdPrognoseGanglinienAnfrage.Daten anfrage;

	/** Die zu erstellenden Prognoseganglinien. */
	private final List<GanglinieInBearbeitung> zyklischeGanglinien;

	/** Flag, ob sich Ereignisse geändert haben. */
	private boolean neueEreignisse;

	/**
	 * Versorgt den Worker mit allen notwendigen Informationen, damit er loslegen kann.
	 *
	 * @param anfrage
	 *            die Anfrage, die der Worker bearbeiten soll.
	 */
	public AnfrageInBearbeitung(final OdPrognoseGanglinienAnfrage.Daten anfrage) {
		zyklischeGanglinien = Collections.synchronizedList(new ArrayList<GanglinieInBearbeitung>());
		this.anfrage = anfrage;
	}

	@Override
	public void run() {
		final long startZeitstempel = System.currentTimeMillis();
		LogTools.log(log, GlProgMsg.InfoStarteAnfrage, anfrage);

		OdPrognoseGanglinienAntwort.Daten datum;
		Iterator<GanglinieInBearbeitung> iterator;

		final ClientDavInterface dav = ObjektFactory.getInstanz().getVerbindung();

		// Auf das Senden der Antwort anmelden
		final OdPrognoseGanglinienAntwort antwort = anfrage.getAbsender()
				.getOnlineDatensatz(OdPrognoseGanglinienAntwort.class);
		try {
			antwort.anmeldenSender(OdPrognoseGanglinienAntwort.Aspekte.Antwort.getAspekt());
		} catch (final AnmeldeException ex) {
			LogTools.log(log, GlProgMsg.ErrorBeimAnmeldenAntwort, anfrage, antwort,
					ex.getLocalizedMessage());
		}
		datum = antwort.erzeugeDatum();
		datum.setAbsenderZeichen(anfrage.getAbsenderZeichen());
		datum.setZeitstempel(anfrage.getZeitstempel());

		// Alle Anfragen abarbeiten und die zyklischen Anfragen
		// zwischenspeichern
		for (final GlProgAnfrage a : anfrage) {
			if ((a.getPrognoseZeitraum() == null) || (a.getPrognoseZeitraum().getLength() == 0)) {
				LogTools.log(log, GlProgMsg.WarningPrognoseIntervallUngueltig, a);
				continue;
			}

			final GanglinieInBearbeitung job = new GanglinieInBearbeitung(a);
			job.startPrognose();
			datum.add(job.getPrognoseganglinie());

			if (a.isZyklischePrognose()) {
				zyklischeGanglinien.add(job);
			}
		}

		// Antwort senden
		try {
			antwort.sendeDaten(OdPrognoseGanglinienAntwort.Aspekte.Antwort.getAspekt(), datum);
		} catch (final DatensendeException ex) {
			LogTools.log(log, GlProgMsg.ErrorBeimAntwortSenden, anfrage, antwort,
					ex.getLocalizedMessage());
		}
		LogTools.log(log, GlProgMsg.InfoAnfrageFertig, anfrage, antwort);
		final long endeZeitstempel = System.currentTimeMillis();
		final long berechnungsdauer = endeZeitstempel - startZeitstempel;
		if ((berechnungsdauer / Constants.MILLIS_PER_SECOND) > 0) {
			LogTools.log(log, GlProgMsg.InfoDauerAnfrageInSekunden,
					berechnungsdauer / Constants.MILLIS_PER_SECOND, anfrage);
		} else {
			LogTools.log(log, GlProgMsg.InfoDauerAnfrageInMillisekunden, berechnungsdauer, anfrage);
		}

		// Falls vorhanden, zyklische Anfragen regelmäßig prüfen
		if (!zyklischeGanglinien.isEmpty()) {
			ObjektFactory.getInstanz().getAOE().addEreignisListener(ereignisListener);
		}
		while (!zyklischeGanglinien.isEmpty()) {
			dav.sleep(Constants.MILLIS_PER_SECOND);

			if (antwort.getStatusSendesteuerung(
					OdPrognoseGanglinienAntwort.Aspekte.Antwort.getAspekt()) != Status.START) {
				LogTools.log(log, GlProgMsg.WarningEmpfaengerAbgemeldet, anfrage);

				// Aufräumen
				for (final GanglinieInBearbeitung g : zyklischeGanglinien) {
					g.close();
				}

				break;
			}

			final long zeit = dav.getTime();
			datum = antwort.erzeugeDatum();
			datum.setAbsenderZeichen(anfrage.getAbsenderZeichen());
			datum.setZeitstempel(zeit);

			iterator = zyklischeGanglinien.iterator();
			while (iterator.hasNext()) {
				final GanglinieInBearbeitung g = iterator.next();
				if (!g.isZyklischAktiv()) {
					g.close();
					iterator.remove();
					continue;
				}

				final boolean aktualisieren = zeit > g.getNaechsteAktualisierung();
				if ((zeit > g.getNaechstePruefung()) || isNeueEreignisse()
						|| g.isPrognoseParameterChanged() || aktualisieren) {
					// Prognoseganglinie muss neu bestimmt werden

					LogTools.log(log, GlProgMsg.FineZyklischPruefen, g.getAnfrage());

					setNeueEreignisse(false);
					final GanglinieMQ letztePrognose = g.getPrognoseganglinie();
					g.startPrognose();
					final int abstand = GanglinienMQOperationen
							.basisabstand(g.getPrognoseganglinie(), letztePrognose);
					LogTools.log(log, GlProgMsg.FinerZyklischAbstand, g.getAnfrage(), abstand);
					if (aktualisieren || (abstand > g.getAnfrage().getSchwelle())) {
						// Prognoseganglinie muss gesendet werden
						LogTools.log(log, GlProgMsg.FineZyklischAktualisieren, g.getAnfrage());
						datum.add(g.getPrognoseganglinie());
					} else {
						LogTools.log(log, GlProgMsg.FineZyklischSchwelleDrunter, g.getAnfrage());
					}
				}
			}

			// Zyklische Prognose senden
			if (!datum.isEmpty()) {
				try {
					antwort.sendeDaten(OdPrognoseGanglinienAntwort.Aspekte.Antwort.getAspekt(),
							datum);
					LogTools.log(log, GlProgMsg.InfoZyklischGesendet, anfrage);
				} catch (final DatensendeException ex) {
					LogTools.log(log, GlProgMsg.ErrorBeimAntwortSenden, anfrage, antwort,
							ex.getLocalizedMessage());
					break;
				}
			}
		}

		ObjektFactory.getInstanz().getAOE().removeEreignisListener(ereignisListener);
		antwort.abmeldenSender(OdPrognoseGanglinienAntwort.Aspekte.Antwort.getAspekt());
	}

	/**
	 * Gibt den Wert des Flags für Ereignisänderung zurück.
	 *
	 * @return {@code true}, wenn sich Ereignisse geändert haben.
	 */
	public synchronized boolean isNeueEreignisse() {
		return neueEreignisse;
	}

	/**
	 * Setzt das Flag für Ereignisänderung.
	 *
	 * @param neueEreignisse
	 *            {@code true}, wenn sich Ereignisse geändert haben.
	 */
	public synchronized void setNeueEreignisse(final boolean neueEreignisse) {
		this.neueEreignisse = neueEreignisse;
	}

}
