/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.glprog.glprog;

import static de.bsvrz.iav.gllib.gllib.dav.GanglinienMQOperationen.addiere;
import static de.bsvrz.iav.gllib.gllib.dav.GanglinienMQOperationen.auschneiden;
import static de.bsvrz.iav.gllib.gllib.dav.GanglinienMQOperationen.multipliziere;
import static de.bsvrz.iav.gllib.gllib.dav.GanglinienMQOperationen.verbinde;
import static de.bsvrz.iav.gllib.gllib.dav.GanglinienMQOperationen.verschiebe;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.bitctrl.Constants;
import com.bitctrl.i18n.Messages;
import com.bitctrl.util.Interval;
import com.bitctrl.util.Shifter;
import com.bitctrl.util.Timestamp;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.iav.gllib.gllib.GanglinienOperationen.PatternMatchingErgebnis;
import de.bsvrz.iav.gllib.gllib.dav.GanglinieMQ;
import de.bsvrz.iav.gllib.gllib.dav.GanglinienMQOperationen;
import de.bsvrz.iav.gllib.gllib.dav.GlProgAnfrage;
import de.bsvrz.iav.gllib.gllib.modell.parameter.PdGanglinienModellAutomatischesLernenEreignis;
import de.bsvrz.iav.gllib.gllib.modell.parameter.PdGanglinienModellPrognose;
import de.bsvrz.iav.glprog.glprog.util.AnalyseganglinieHelper;
import de.bsvrz.iav.glprog.glprog.util.EreignisHelper;
import de.bsvrz.iav.glprog.glprog.util.HistorischeGanglinienHelper;
import de.bsvrz.iav.glprog.glprog.util.PrognoseParameterHelper;
import de.bsvrz.sys.funclib.bitctrl.daf.BetriebsmeldungDaten;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.kalender.objekte.Ereignis;
import de.bsvrz.sys.funclib.bitctrl.modell.kalender.objekte.EreignisTyp;
import de.bsvrz.sys.funclib.bitctrl.modell.kalender.onlinedaten.OdEreignisKalenderAntwort;
import de.bsvrz.sys.funclib.bitctrl.modell.kalender.onlinedaten.OdEreignisKalenderAntwort.Daten.Zustandswechsel;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Kapselt alle Daten, die zur Bearbeitung einer Anfrage notwendig sind.
 *
 * @author BitCtrl Systems GmbH, Schumann
 */
public class GanglinieInBearbeitung {

	/** Der Logger der Klasse. */
	private final Debug log = Debug.getLogger();

	/** Die zu bearbeitende Anfrage. */
	private final GlProgAnfrage anfrage;

	/** Das Zeichen des Absenders der Prognoseanfrage. */
	private final String absenderZeichen;

	/** Die historischen Ganglinien am Messquerschnitt. */
	private List<GanglinieMQ> historischeGanglinien;

	/** Die Prognoseparameter. */
	private PdGanglinienModellPrognose.Daten prognoseParameter;

	/** Die Antwort vom Kalender. */
	private List<Zustandswechsel> kalenderantwort;

	/** Die Analyseganglinie, die aus dem Archiv gelesen wird. */
	private GanglinieMQ analyseganglinie;

	/** Enthält die Teilintervalle des Prognosezeitraums und deren Ganglinien. */
	private List<PrognoseIntervall> prognoseintervalle;

	/** Die prognostizierte Ganglinie. */
	private GanglinieMQ prognoseganglinie;

	/** Flag, ob sich der Prognoseparameter am Messquerschnitt geändert hat. */
	private boolean prognoseParameterChanged;

	/** Zeitstempel, wann die nächste Prüfung der Prognose fällig ist. */
	private long naechstePruefung;

	/** Zeitstempel, wann die nächste Aktualisierung der Prognose fällig ist. */
	private long naechsteAktualisierung;

	private final PrognoseParameterHelper prognoseParameterHelper;

	private final PropertyChangeListener prognoseParameterListener = new PropertyChangeListener() {

		@Override
		public void propertyChange(final PropertyChangeEvent evt) {
			setPrognoseParameterChanged(true);
		}

	};

	/**
	 * Initialisierung die Bearbeitung der Prognoseanfrage.
	 *
	 * @param anfrage
	 *            die zu bearbeitende Anfrage.
	 */
	public GanglinieInBearbeitung(final GlProgAnfrage anfrage) {
		final ClientDavInterface dav = ObjektFactory.getInstanz().getVerbindung();
		this.anfrage = anfrage;

		absenderZeichen = anfrage.getMessQuerschnitt() + "/" + dav.getTime(); //$NON-NLS-1$

		prognoseParameterHelper = new PrognoseParameterHelper();
		prognoseParameterHelper.setAnfrage(anfrage);
		prognoseParameter = prognoseParameterHelper.getPrognoseParameter();
		if (anfrage.isZyklischePrognose()) {
			prognoseParameterHelper.addPropertyChangeListener(prognoseParameterListener);
		}
	}

	/**
	 * Schließt alle offenen Ressourcen. Das Objekt darf danach nicht mehr verwendet werden.
	 */
	public void close() {
		if (anfrage.isZyklischePrognose()) {
			prognoseParameterHelper.removePropertyChangeListener(prognoseParameterListener);
		}
	}

	/**
	 * Gibt die zu bearbeitende Anfrage zurück.
	 *
	 * @return die Anfrage.
	 */
	public GlProgAnfrage getAnfrage() {
		return anfrage;
	}

	/**
	 * Gibt die Prognoseganglinie zurück.
	 *
	 * @return eine Ganglinie, niemals {@code null}.
	 */
	public GanglinieMQ getPrognoseganglinie() {
		if (prognoseganglinie != null) {
			return prognoseganglinie;
		}

		// Leere Ganglinie erstellen, wenn keine Prognose möglich war
		final GanglinieMQ ganglinie = new GanglinieMQ();
		ganglinie.setMessQuerschnitt(getAnfrage().getMessQuerschnitt());
		if (getAnfrage().getPrognoseZeitraum() != null) {
			ganglinie.setPrognoseZeitraum(getAnfrage().getPrognoseZeitraum());
		} else {
			ganglinie.setPrognoseZeitraum(
					new Interval(System.currentTimeMillis(), System.currentTimeMillis()));
		}
		return ganglinie;
	}

	/**
	 * Startet die Bearbeitung.
	 */
	public void startPrognose() {
		final long startZeitstempel = System.currentTimeMillis();
		long start;
		boolean ok = true;

		start = ObjektFactory.getInstanz().getVerbindung().getTime();

		LogTools.log(log, GlProgMsg.FineStartBearbeitungGanglinie, anfrage);

		// Eingangsdaten ermitteln
		bestimmeEingangsdaten();
		setPrognoseParameterChanged(false);
		LogTools.log(log, GlProgMsg.FineEingangsdatenKomplett, anfrage);

		if (historischeGanglinien.isEmpty()) {
			ok = false;
			LogTools.log(log,
					new BetriebsmeldungDaten(anfrage.getMessQuerschnitt().getSystemObject()),
					GlProgMsg.WarningKeineGanglinien, anfrage.getMessQuerschnitt());
		}

		if ((kalenderantwort == null) || kalenderantwort.isEmpty()) {
			ok = false;
			LogTools.log(log,
					new BetriebsmeldungDaten(anfrage.getMessQuerschnitt().getSystemObject()),
					GlProgMsg.WarningKalenderKeineEreignisse, anfrage.getMessQuerschnitt(),
					anfrage.getPrognoseZeitraum());
		}

		if (!prognoseParameter.isValid()) {
			ok = false;
			LogTools.log(log,
					new BetriebsmeldungDaten(anfrage.getMessQuerschnitt().getSystemObject()),
					GlProgMsg.WarningPrognoseparameterNichtGesetzt, anfrage.getMessQuerschnitt(),
					PdGanglinienModellPrognose.ATG_GANGLINIEN_MODELL_PROGNOSE);
		}

		if (!ok) {
			return;
		}

		// Prognose vorbereiten
		aufbereitenZustandswechsel();
		bestimmeTeilintervalle();

		if ((prognoseintervalle == null) || prognoseintervalle.isEmpty()) {
			prognoseganglinie = null;
			LogTools.log(log, GlProgMsg.WarningKeineEreignisse, anfrage);
			return;
		}

		// Prognose durchführen
		bestimmeTagesganglinien();
		bestimmeErlaubteRelativeEreignisTypen();
		bestimmePrognoseGanglinien();
		prognoseganglinie = erzeugePrognoseganglinie();
		if (prognoseganglinie != null) {
			prognoseganglinie.setPrognoseZeitraum(anfrage.getPrognoseZeitraum());
		}
		LogTools.log(log, GlProgMsg.FinerPrognoseganglinie, anfrage, prognoseganglinie);

		historischeGanglinien = null;
		kalenderantwort = null;
		prognoseintervalle = null;
		prognoseParameter = null;
		naechstePruefung = start + anfrage.getPruefIntervall();
		naechsteAktualisierung = start + anfrage.getSendeIntervall();
		LogTools.log(log, GlProgMsg.FineGangliniePrognostiziert, anfrage, prognoseganglinie);

		final long endeZeitstempel = System.currentTimeMillis();
		final long berechnungsdauer = endeZeitstempel - startZeitstempel;
		if ((berechnungsdauer / Constants.MILLIS_PER_SECOND) > 0) {
			LogTools.log(log,
					new BetriebsmeldungDaten(anfrage.getMessQuerschnitt().getSystemObject()),
					GlProgMsg.FineDauerGanglinienanfrageInSekunden,
					berechnungsdauer / Constants.MILLIS_PER_SECOND, anfrage);
		} else {
			LogTools.log(log,
					new BetriebsmeldungDaten(anfrage.getMessQuerschnitt().getSystemObject()),
					GlProgMsg.FineDauerGanglinienanfrageInMillisekunden, berechnungsdauer, anfrage);
		}
	}

	/**
	 * Gibt den nächsten Zeitpunkt zurück, an dem die Prognose überprüft werden muss.
	 *
	 * @return der Zeitstempel mit der Fälligkeit der Prognoseprüfung.
	 */
	public long getNaechstePruefung() {
		return naechstePruefung;
	}

	/**
	 * Gibt den nächsten Zeitpunkt zurück, an dem die Prognose aktualisiert werden muss.
	 *
	 * @return der Zeitstempel mit der Fälligkeit der Prognoseaktualisierung.
	 */
	public long getNaechsteAktualisierung() {
		return naechsteAktualisierung;
	}

	/**
	 * Bereitet die Liste der Zustandswechsel auf, damit der Scanline-Algorithmus aus
	 * {@link #bestimmeTeilintervalle()} funktioniert.
	 * <p>
	 * Am Anfang der Liste können Zustandswechsel fehlen, die den Beginn des Ereignisses
	 * signalisieren, wenn das Ereignis vor Beginn des Anfragezeitraumes begonnen hat.
	 * <p>
	 * Ereignisse die vor dem Anfragezeitraum enden, werden aus der Liste entfernt.
	 * <p>
	 * Da die Gültigkeit für Ereignisse zeitlich und verkehrlich unterschiedlich ist, hier aber nur
	 * die verkehrliche Gültigkeit interessiert, werden unnötige Zustandswechsel entfernt. So ein
	 * unnötiger Zustandswechsel ist z.&nbsp;B. wenn zweimal auf einander der selbe Zustand für
	 * verkehrliche Gültigkeit kommt, weil sich die zeitliche Gültigkeit geändert hat.
	 */
	private void aufbereitenZustandswechsel() {
		Iterator<OdEreignisKalenderAntwort.Daten.Zustandswechsel> iterator;
		final Set<Ereignis> scanline;
		Set<OdEreignisKalenderAntwort.Daten.Zustandswechsel> beginnen;

		/*
		 * Für endende, aber nicht beginnende Ereignisse, den Anfangszeitpunkt auf Prognosebeginn
		 * setzen.
		 */
		scanline = new HashSet<>();
		beginnen = new HashSet<>();
		iterator = kalenderantwort.iterator();
		while (iterator.hasNext()) {
			OdEreignisKalenderAntwort.Daten.Zustandswechsel z;

			z = iterator.next();
			if (z.isVerkehrlichGueltig()) {
				// Ereignis beginnt
				if (z.getZeitstempel() >= anfrage.getPrognoseZeitraum().getEnd()) {
					/*
					 * Ereignis beginnt am Ende oder nach dem Prognoseintervall und kann deswegen
					 * ignoriert werden.
					 */
					iterator.remove();
				} else {
					scanline.add(z.getEreignis());
				}
			} else {
				if (scanline.contains(z.getEreignis())) {
					// begonnenes Ereignis endet
					scanline.remove(z.getEreignis());
				} else {
					// Ereignis endet, ohne begonnen zu haben
					if (z.getZeitstempel() <= anfrage.getPrognoseZeitraum().getStart()) {
						/*
						 * Ereignis endet vor dem Prognoseintervall und kann komplett ignoriert
						 * werden.
						 */
						iterator.remove();
					} else {
						// Den Beginn des Ereignisses ergänzen
						beginnen.add(new OdEreignisKalenderAntwort.Daten.Zustandswechsel(
								anfrage.getPrognoseZeitraum().getStart(), z.getEreignis(), true,
								true));
					}
				}
				if (z.getZeitstempel() >= anfrage.getPrognoseZeitraum().getEnd()) {
					/*
					 * Ereignis endet am Ende oder nach dem Prognoseintervall, der Zustandswechsel
					 * ist nicht relevant
					 */
					iterator.remove();
				}
			}
		}
		for (final OdEreignisKalenderAntwort.Daten.Zustandswechsel z : beginnen) {
			kalenderantwort.add(0, z);
		}

		// Doppelte Zustandswechsel aus der Liste entfernen.
		scanline.clear();
		iterator = kalenderantwort.iterator();
		while (iterator.hasNext()) {
			OdEreignisKalenderAntwort.Daten.Zustandswechsel z;

			z = iterator.next();
			if (z.isVerkehrlichGueltig()) {
				if (scanline.contains(z.getEreignis())) {
					// Doppelter Beginn
					iterator.remove();
				} else {
					scanline.add(z.getEreignis());
				}
			} else {
				if (scanline.contains(z.getEreignis())) {
					scanline.remove(z.getEreignis());
				} else {
					// Doppeltes Ende
					iterator.remove();
				}
			}
		}
	}

	/**
	 * Erstellt aus der Liste der Zustandswechsels aus der Kalenderantwort die enhtsprechenden
	 * Ereignisintervalle. Es wird ein Scanline-Algorithmus verwendet, der die Liste der
	 * Zustandswechsel "scannt".
	 * <p>
	 * Die Methode entspricht zum Teil der ereignistypabhängigen Auswahl von Ganglinien laut
	 * Feinspezifikation. Da der Ereigniskalender in seiner Antwort nicht zwischen Tagesereignissen
	 * und Nichttagesereignissen unterscheidet, wird diese Auswahl sowohl für relative als auch für
	 * absolute Ganglinien durchgeführt.
	 * <p>
	 * Da die priorisierte Auswahl von Tagesganglinien und die ereignistypabhängige Auswahl von
	 * (relativen) Ganglinien voneinander unabhängig sind, stellt die gleichzeitige Ausführung keine
	 * Verletzung des Algorithmus dar.
	 * <p>
	 * Am Ende der Methode ist ist die Liste der Teilintervalle gefüllt (falls möglich).
	 *
	 * @see #prognoseintervalle
	 */
	private void bestimmeTeilintervalle() {
		Map<EreignisTyp, EreignisTypGanglinien> ganglinien;
		Set<Ereignis> scanline;
		long start = 0;

		ganglinien = pruefeGanglinien(historischeGanglinien);
		prognoseintervalle = new ArrayList<>();

		// Die Liste der Ereignisytpen im aktuellen Abschnitt
		scanline = new HashSet<>();

		/*
		 * Scanline auf der Zeitachse entlangschieben. Die Scanline wird immer beim Tageswechsel
		 * anhalten, da angenommen wird, dass die Ereignisse Montag, Dienstag usw. immer existieren.
		 */
		for (final OdEreignisKalenderAntwort.Daten.Zustandswechsel z : kalenderantwort) {
			Ereignis ereignis;

			ereignis = z.getEreignis();

			if (scanline.isEmpty()) {
				// Dies ist der Beginn eines neuen Intervalls
				start = z.getZeitstempel();
				scanline.add(ereignis);
				continue;
			}

			if ((start == z.getZeitstempel()) && z.isVerkehrlichGueltig()) {
				/*
				 * Das Ereignis beginnt zusammen mit denen in der Scanline, ergo kein neues
				 * Intervall notwendig.
				 */
				scanline.add(ereignis);
			} else if ((start == z.getZeitstempel()) && !z.isVerkehrlichGueltig()) {
				/*
				 * Das Ereignis endet zu Beginn des Scanline-Abschnitts. Der Ereignistyp wird nur
				 * aus der Scanline entfernt, aber kein neues Intervall begonnnen, da es die Breite
				 * 0 hätte.
				 */
				scanline.remove(ereignis);
			} else if (start != z.getZeitstempel()) {
				assert start <= z
						.getZeitstempel() : "Die zeitliche Reihenfolge der Zustände ist nicht fortlaufend."; //$NON-NLS-1$

						/*
						 * Ein Ereignis beginnt oder endet innerhalb eines anderen. Es wird die Scanline als
						 * Teilintervall abgelegt.
						 */
						if (z.getZeitstempel() > anfrage.getPrognoseZeitraum().getStart()) {
							final PrognoseIntervall progIntervall;
							final long start1;

							if (start < anfrage.getPrognoseZeitraum().getStart()) {
								/*
								 * start liegt vor dem Prognosezeitraum und muss korrigiert werden
								 */
								start1 = anfrage.getPrognoseZeitraum().getStart();
							} else {
								start1 = start;
							}
							progIntervall = new PrognoseIntervall(new Interval(start1, z.getZeitstempel()));
							for (final Ereignis e : scanline) {
								if (ganglinien.containsKey(e.getEreignisTyp())) {
									progIntervall.add(ganglinien.get(e.getEreignisTyp()).clone());
								} else {
									LogTools.log(log,
											new BetriebsmeldungDaten(
													anfrage.getMessQuerschnitt().getSystemObject()),
													GlProgMsg.WarningKeineGanglinieFuerEreignistyp,
													anfrage.getMessQuerschnitt(), e.getEreignisTyp());
								}
							}
							prognoseintervalle.add(progIntervall);
						}
						if (z.isVerkehrlichGueltig()) {
							// Ein neues Ereignis beginnt
							scanline.add(ereignis);
						} else {
							// Ein Ereignis der Scanline endet
							scanline.remove(ereignis);
						}
						start = z.getZeitstempel();
			} else {
				assert false : "Unbehandelter Zustandswechsel"; //$NON-NLS-1$
			}
		}

		/*
		 * Falls die Scanline noch Ereignisse enthält, wird ein Intervall bis zum Ende des
		 * Prognosezeitraums ergänzt.
		 */
		if (!scanline.isEmpty()) {
			PrognoseIntervall progIntervall;
			final long start1;

			if (start < anfrage.getPrognoseZeitraum().getStart()) {
				/*
				 * start liegt vor dem Prognosezeitraum und muss korrigiert werden
				 */
				start1 = anfrage.getPrognoseZeitraum().getStart();
			} else {
				start1 = start;
			}
			progIntervall = new PrognoseIntervall(
					new Interval(start1, anfrage.getPrognoseZeitraum().getEnd()));
			for (final Ereignis e : scanline) {
				if (ganglinien.containsKey(e.getEreignisTyp())) {
					progIntervall.add(ganglinien.get(e.getEreignisTyp()).clone());
				} else {
					LogTools.log(log, GlProgMsg.WarningKeineGanglinieFuerEreignistyp,
							anfrage.getMessQuerschnitt(), e.getEreignisTyp());
				}
			}
			prognoseintervalle.add(progIntervall);
		}

		LogTools.log(log, GlProgMsg.FinerPrognoseintervalle, anfrage, prognoseintervalle);
	}

	/**
	 * Führt die priorisierte Auswahl der Tagesganglinie pro Tag durch. Als Ergebnis enthält die
	 * Liste der Teilintervalle pro Teilintervall nur noch ein Tagesereignis. Die Teilintervalle
	 * eines Tages besitzen alle das selbe Tagesereignis.
	 * <p>
	 * Die Methode setzt voraus, dass die Teilintervalle des Prognosezeitraums bestimmt wurde.
	 *
	 * @see #bestimmeTeilintervalle()
	 */
	private void bestimmeTagesganglinien() {
		EreignisTyp tagestyp = null;

		for (final PrognoseIntervall intervall : prognoseintervalle) {
			List<EreignisTyp> ereignisTypen;
			Iterator<EreignisTypGanglinien> iterator;
			Calendar kalender;

			ereignisTypen = intervall.getEreignisTypenAbsolut();
			if (ereignisTypen.isEmpty()) {
				/*
				 * Es gibt keine absoluten Ganglinien im Teilintervall und kann deswegen nicht
				 * weiter betrachtet werden.
				 */
				intervall.clear();
				continue;
			}

			kalender = Calendar.getInstance();
			kalender.setTimeInMillis(intervall.getIntervall().getStart());
			if ((kalender.get(Calendar.HOUR_OF_DAY) == 0) && (kalender.get(Calendar.MINUTE) == 0)
					&& (kalender.get(Calendar.SECOND) == 0)
					&& (kalender.get(Calendar.MILLISECOND) == 0)) {
				// Intervall ist der Beginn eines neuen Tages
				tagestyp = null;
			}

			if (tagestyp == null) {
				/*
				 * Neuer aktueller Tagesereignistyp wird festgelegt und notwendige Parameter
				 * abgerufen.
				 */
				tagestyp = ereignisTypen.get(0);
				tagestyp.getParameterDatensatz(PdGanglinienModellAutomatischesLernenEreignis.class)
				.abrufenDatum();
			}
			assert ereignisTypen.contains(tagestyp);
			while (ereignisTypen.contains(tagestyp)) {
				ereignisTypen.remove(tagestyp);
			}

			// Alle anderen Tagesganglinien im Teilintervall entfernen
			iterator = intervall.iterator();
			while (iterator.hasNext()) {
				EreignisTypGanglinien ganglinien;

				ganglinien = iterator.next();
				if (ereignisTypen.contains(ganglinien.getEreignisTyp())) {
					iterator.remove();
				}
			}
		}

		LogTools.log(log, GlProgMsg.FinerPrognoseintervalleNachPriorisierteAuswahl, anfrage,
				prognoseintervalle);
	}

	/**
	 * Entfernt pro Teilintervall der Prognose alle Ereignistypen, die sich nicht mit dem gewählten
	 * Tagesereignis verknüpfen lassen.
	 * <p>
	 * Die Methode setzt voraus, dass es nur noch ein Tagesereignis pro Teilintervall der Prognose
	 * gibt.
	 *
	 * @see #bestimmeTagesganglinien()
	 */
	private void bestimmeErlaubteRelativeEreignisTypen() {
		for (final PrognoseIntervall intervall : prognoseintervalle) {
			EreignisTypGanglinien tagesGl;
			Iterator<EreignisTypGanglinien> iterator;

			tagesGl = intervall.getTagesEreignisTypGanglinien();
			if (tagesGl == null) {
				// Im aktuellen Teilintervall gibt es keine Ganglinien.
				continue;
			}

			iterator = intervall.getEreignisTypGanglinienRelativ().iterator();
			while (iterator.hasNext()) {
				EreignisTypGanglinien relativeGl;
				PdGanglinienModellAutomatischesLernenEreignis.Daten datum;

				/*
				 * Verwerfe alle relativen Ganglinien, die sich nicht auf das ausgewählte
				 * Tagesereignis beziehen können.
				 */
				relativeGl = iterator.next();
				datum = relativeGl.getEreignisTyp()
						.getParameterDatensatz(PdGanglinienModellAutomatischesLernenEreignis.class)
						.abrufenDatum();

				if (datum.isValid()) {
					List<EreignisTyp> bezug;

					bezug = datum.getBezugsereignistypen();

					if (!bezug.isEmpty() && !bezug.contains(tagesGl.getEreignisTyp())) {
						iterator.remove();
					}
				}
			}
		}

		LogTools.log(log, GlProgMsg.FinerPrognoseintervalleNachEntfernenEreignisse, anfrage,
				prognoseintervalle);
	}

	/**
	 * Bestimmt eine relative Ganglinie pro Ereignistyp, die zur Prognose herangezogen wird. Im
	 * Zeitbereich vor dem Horizont der mittelfristigen Prognose wird das Pattern-Matching zur
	 * Auswahl verwendet. Danach werden die Auswahlverfahren der langfristigen Prognose
	 * (statistische Auswahl und Referenzganglinie) verwendet. Die Methoden der langfristigen
	 * Prognose werden auch angewandt, wenn das Pattern-Matching kein Ergebnis geliefert hat.
	 * <p>
	 * Ist der Horizont der mittelfristigen Prognose gleich 0, werden nur die Auswahlmethode der
	 * langfristigen Prognose angewandt.
	 */
	private void bestimmePrognoseGanglinien() {
		for (final PrognoseIntervall intervall : prognoseintervalle) {
			long mittelfristigePrognose, horizont;
			EreignisTypGanglinien tagesGl;
			boolean patternMatching = false;

			tagesGl = intervall.getTagesEreignisTypGanglinien();
			if (tagesGl == null) {
				// Im aktuellen Teilintervall gibt es keine Ganglinien.
				continue;
			}

			if (anfrage.isNurLangfristigeAuswahl()) {
				mittelfristigePrognose = 0;
			} else {
				mittelfristigePrognose = prognoseParameter.getPatternMatchingHorizont();
			}
			assert mittelfristigePrognose >= 0;

			horizont = intervall.getIntervall().getStart()
					- anfrage.getPrognoseZeitraum().getStart();

			// Versuche mittelfristige Prognose
			if ((analyseganglinie != null) && (analyseganglinie.size() > 0) && (0 <= horizont)
					&& (horizont < mittelfristigePrognose)) {
				// Mittelfristige Prognose
				patternMatching = patternMatching(intervall);
			}

			if ((horizont >= mittelfristigePrognose) || !patternMatching) {

				for (final EreignisTypGanglinien ergGl : intervall) {
					// Mache langfristige Prognose Langfristige Prognose
					if (prognoseParameter
							.getAuswahlMethode() == PdGanglinienModellPrognose.Daten.REFERENZGANGLINIE) {
						/*
						 * Auswahl nach Referenzganglinie. Es werden alle Ganglinien entfernt, die
						 * keine Referenzganglinie sind.
						 */
						Iterator<GanglinieMQ> gl;

						gl = ergGl.iterator();
						while (gl.hasNext()) {
							if (!gl.next().isReferenz()) {
								gl.remove();
							}
						}

						/*
						 * Falls mehr als eine Referenzganglinie vorhanden ist, eine auswählen und
						 * Warnung ausgeben.
						 */
						if (ergGl.size() > 1) {
							LogTools.log(log,
									new BetriebsmeldungDaten(
											anfrage.getMessQuerschnitt().getSystemObject()),
											GlProgMsg.WarningMehrAlsEineReferenzganglinie,
											anfrage.getMessQuerschnitt(), ergGl.getEreignisTyp());
						}
						while (ergGl.size() > 1) {
							ergGl.remove(0);
						}
					} else {
						/*
						 * Auswahl nach statistischer Methode, ergo Auswahl der Ganglinie mit der
						 * höchsten Anzahl der Verschmelzungen
						 */

						// Aufsteigend nach Anzahl der Verschmelzungen sortieren
						Collections.sort(ergGl, new Comparator<GanglinieMQ>() {

							@Override
							public int compare(final GanglinieMQ g1, final GanglinieMQ g2) {
								if (g1.getAnzahlVerschmelzungen() < g2.getAnzahlVerschmelzungen()) {
									return -1;
								} else if (g1.getAnzahlVerschmelzungen() > g2
										.getAnzahlVerschmelzungen()) {
									return 1;
								}
								return 0;
							}

						});

						/*
						 * Alle Ganglinien entfernen, bis auf die letzte, die die mit der höchsten
						 * Anzahl der Verschmelzungen ist.
						 */
						while (ergGl.size() > 1) {
							ergGl.remove(0);
						}
					}
				}
			}
		}

		LogTools.log(log, GlProgMsg.FinerPrognoseintervalleNachMittelUndLangFristigerPrognose,
				anfrage, prognoseintervalle);
	}

	/**
	 * Führt das Pattern-Matching für ein Teilintervall aus.
	 *
	 * @param intervall
	 *            das Intrvall für das das Pattern-Matching ausgeführt werden soll.
	 * @return {@code true}, wenn das Pattern Matching erfolgreich war.
	 */
	private boolean patternMatching(final PrognoseIntervall intervall) {
		LogTools.log(log, GlProgMsg.FinerPatternMatchingFuerPrognoseIntervall, anfrage, intervall);

		final List<GanglinieMQ> listeMatching, listeAbsolut;
		final List<EreignisTypGanglinien> listeRelativ;

		listeAbsolut = intervall.getTagesEreignisTypGanglinien();
		listeRelativ = intervall.getEreignisTypGanglinienRelativ();

		if (listeAbsolut.isEmpty()) {
			throw new IllegalStateException(Messages.get(
					GlProgMsg.IllegalStateExceptionPrognoseIntervallEnthaeltKeineTagesganglinien,
					intervall));
		}

		// Vergleichsliste zusammenstellen
		listeMatching = new ArrayList<>();
		for (final GanglinieMQ glAbsolut : listeAbsolut) {
			GanglinieMQ ganglinie;

			if (listeRelativ.size() > 0) {
				final Shifter shifter;
				final int[] maximas;

				/*
				 * Der Shifter hält pro relativen Ereignis den Index der aktuellen Ganglinie
				 */
				maximas = new int[listeRelativ.size()];
				for (int i = 0; i < maximas.length; ++i) {
					maximas[0] = listeRelativ.get(i).size() - 1;
				}
				shifter = new Shifter(maximas);

				do {
					final List<GanglinieMQ> listeAdd, listeMul;

					/*
					 * Bestimme für jedes relative Ereignis die aktuelle Ganglinie und füge sie der
					 * Liste der additiven bzw. multiplikativen Ganglinien hinzu
					 */
					listeAdd = new ArrayList<>();
					listeMul = new ArrayList<>();
					for (int i = 0; i < shifter.getLength(); ++i) {
						GanglinieMQ g;

						g = listeRelativ.get(i).get(shifter.getValues()[i]);
						if (g.getTyp() == GanglinieMQ.TYP_ADDITIV) {
							listeAdd.add(g);
						} else if (g.getTyp() == GanglinieMQ.TYP_MULTIPLIKATIV) {
							listeMul.add(g);
						} else {
							throw new IllegalStateException(Messages.get(
									GlProgMsg.IllegalStateExceptionFalscherGanglinienTyp,
									g.getTyp()));
						}
					}

					// Kombiniere die absolute mit den relativen Ganglinien
					ganglinie = glAbsolut.clone();
					for (final GanglinieMQ g : listeMul) {
						if (ganglinie.getApproximationDaK() == g.getApproximationDaK()) {
							ganglinie = multipliziere(ganglinie, g);
						}
						// Bei unterschiedlicher Approximation, Ganglinie ignorieren
					}
					for (final GanglinieMQ g : listeAdd) {
						if (ganglinie.getApproximationDaK() == g.getApproximationDaK()) {
							ganglinie = addiere(ganglinie, g);
						}
						// Bei unterschiedlicher Approximation, Ganglinie ignorieren
					}

					// Die bestimmte Ganglinien in Vergleichsliste aufnehmen
					listeMatching.add(ganglinie);
				} while (shifter.shift());
			} else {
				listeMatching.add(glAbsolut.clone());
			}
		}

		/*
		 * Archivganglinie auf 1.1.1970 verschieben, damit sie mit historischen Ganglinien
		 * verglichen werden kann.
		 */
		final GanglinieMQ archivGl;
		final Calendar cal;
		final EreignisTypGanglinien result;
		final PatternMatchingErgebnis<GanglinieMQ> matchingErg;

		archivGl = analyseganglinie.clone();
		cal = Calendar.getInstance();
		cal.setTimeInMillis(archivGl.getStuetzstellen().get(0).getZeitstempel());
		cal.set(Calendar.HOUR_OF_DAY, 0);
		cal.set(Calendar.MINUTE, 0);
		cal.set(Calendar.SECOND, 0);
		cal.set(Calendar.MILLISECOND, 0);
		GanglinienMQOperationen.verschiebe(archivGl, -cal.getTimeInMillis());

		// Eigentliches Pattern-Matching ausführen
		matchingErg = GanglinienMQOperationen.patternMatching(archivGl, listeMatching,
				prognoseParameter.getPatternMatchingOffset(),
				prognoseParameter.getPatternMatchingOffset(),
				prognoseParameter.getMatchingIntervall());
		LogTools.log(log, GlProgMsg.FinerErgebnisPatternMatching, anfrage, matchingErg);

		if (matchingErg.getAbstand() > prognoseParameter.getMaxMatchingFehler()) {
			// Keine passende Ganglinie gefunden
			return false;
		}

		// Alle anderen Ganglinie im Intervall entfernen und Ergebnis sichern
		result = new EreignisTypGanglinien(listeAbsolut.get(0).getEreignisTyp());
		result.setTagesganglinie(true);
		result.add(matchingErg.getGanglinie());
		intervall.clear();
		intervall.add(result);
		return true;
	}

	/**
	 * Prüft ob die Liste der Ganglinien pro Ereignistyp plausibel ist. Pro Ereignistyp darf es
	 * entweder nur absolute oder nur relative Ganglinien geben. Ist dies nicht der Fall, ist die
	 * Ganglinienliste für das betreffende Ereignis ungültig und wird ignoriert. Für den
	 * betreffenden Ganglinientyp wird eine Warnung ausgegeben.
	 *
	 * @param ganglinien
	 *            alle verf&uuml;gbaren Ganglinien im Prognosezeitraum.
	 * @return eine {@code Map} mit g&uuml;ltigen Ganglinien pro Ereignistyp.
	 */
	private Map<EreignisTyp, EreignisTypGanglinien> pruefeGanglinien(
			final List<GanglinieMQ> ganglinien) {
		final Map<EreignisTyp, EreignisTypGanglinien> result;
		final Set<EreignisTyp> ausschuss;

		result = new HashMap<>();
		ausschuss = new HashSet<>();
		// Alle Ganglinien durchlaufen und nach Ereignistyp in Ergebis sortieren
		for (final GanglinieMQ gl : ganglinien) {
			EreignisTyp glTyp;
			EreignisTypGanglinien glListe;

			glTyp = gl.getEreignisTyp();
			if (ausschuss.contains(glTyp)) {
				/*
				 * Der Ereignistyp dieser Ganglinie wurde bereits als ungültig erkannt, ergo muss
				 * die Ganglinie nicht weiter untersucht werden.
				 */
				continue;
			}

			if (!result.containsKey(glTyp)) {
				/*
				 * Der Ereignistyp wird zum ersten Mal untersucht, ergo erstmal in Ergebnis mit
				 * leerer Ganglinienliste aufnehmen.
				 */
				result.put(glTyp, new EreignisTypGanglinien(glTyp));
			}
			glListe = result.get(glTyp);

			if (glListe.size() == 0) {
				/*
				 * Dies ist die erste Ganglinie des Ereignistyps, weswegen er den bestimmt ob die
				 * Ganglinien zu einem Tagesereignis gehören oder nicht.
				 */
				glListe.add(gl);
				glListe.setTagesganglinie(gl.isAbsolut());
				continue;
			}

			if ((glListe.isTagesganglinie() && gl.isAbsolut())
					|| (!glListe.isTagesganglinie() && !gl.isAbsolut())) {
				/*
				 * Nach hinzufügen der Ganglinie würde die Liste des Ereignistyps nur absolute oder
				 * nur relative Ganglinien / enthalten.
				 */
				glListe.add(gl);
			} else {
				/*
				 * Es gibt für den Ereignistyp absolute und relative Ganglinien. Dies ist ein
				 * inkonsistenter Zustand.
				 */
				ausschuss.add(glTyp);
				result.remove(glTyp);
				LogTools.log(log,
						new BetriebsmeldungDaten(anfrage.getMessQuerschnitt().getSystemObject()),
						GlProgMsg.WarningGanglinienInkonistent, anfrage.getMessQuerschnitt(),
						glTyp);
			}
		}

		return result;
	}

	/**
	 * Verschmilzt pro Prognoseintervall die realtiven Ganglinien mit der absoluten Ganglinie.
	 *
	 * @return die Prognoseganglinie, f&uuml;r das gesamte Prognoseintervall.
	 */
	private GanglinieMQ erzeugePrognoseganglinie() {
		List<GanglinieMQ> teile;
		GanglinieMQ ganglinie = null;
		Calendar tag;

		// Ganglinie pro Teilintervall bestimmen
		tag = Calendar.getInstance();
		tag.setTimeInMillis(anfrage.getPrognoseZeitraum().getStart());
		tag.set(Calendar.HOUR_OF_DAY, 0);
		tag.set(Calendar.MINUTE, 0);
		tag.set(Calendar.SECOND, 0);
		tag.set(Calendar.MILLISECOND, 0);
		teile = new ArrayList<>();
		for (final PrognoseIntervall intervall : prognoseintervalle) {
			GanglinieMQ intervallGl;
			List<GanglinieMQ> addGl;
			List<GanglinieMQ> mulGl;
			Calendar teilIntervall;

			teilIntervall = Calendar.getInstance();
			teilIntervall.setTimeInMillis(intervall.getIntervall().getStart());
			if ((teilIntervall.get(Calendar.HOUR_OF_DAY) == 0)
					&& (teilIntervall.get(Calendar.MINUTE) == 0)
					&& (teilIntervall.get(Calendar.SECOND) == 0)
					&& (teilIntervall.get(Calendar.MILLISECOND) == 0)) {
				tag = teilIntervall;
			}

			if ((intervall.getTagesEreignisTypGanglinien() == null)
					|| intervall.getTagesEreignisTypGanglinien().isEmpty()) {
				// Es gibt keine Ganglinien im Teilintervall
				continue;
			}

			// Tagesganglinie im Teilintervall bestimmen
			assert intervall.getTagesEreignisTypGanglinien()
			.size() == 1 : "Es darf wegen der vorangegangenen Auswahl nur noch eine Tagesganglinie geben."; //$NON-NLS-1$
			intervallGl = intervall.getTagesEreignisTypGanglinien().get(0).clone();

			// Additive relative Ganglinien bestimmen
			addGl = new ArrayList<>();
			for (final EreignisTypGanglinien typGl : intervall.getEreignisTypGanglinienRelativ()) {
				if (!typGl.isEmpty() && (typGl.get(0).getTyp() == GanglinieMQ.TYP_ADDITIV)) {
					addGl.add(typGl.get(0));
				}
			}

			// Multiplikative relative Ganglinien bestimmen
			mulGl = new ArrayList<>();
			for (final EreignisTypGanglinien typGl : intervall.getEreignisTypGanglinienRelativ()) {
				if (!typGl.isEmpty() && (typGl.get(0).getTyp() == GanglinieMQ.TYP_MULTIPLIKATIV)) {
					mulGl.add(typGl.get(0));
				}
			}

			// Tagesganglinie und relative Ganglinien verknüpfen
			for (final GanglinieMQ g : mulGl) {
				if (intervallGl.getApproximationDaK() == g.getApproximationDaK()) {
					intervallGl = multipliziere(intervallGl, g);
				} // Bei unterschiedlicher Approximation, Ganglinie ignorieren
			}
			for (final GanglinieMQ g : addGl) {
				if (intervallGl.getApproximationDaK() == g.getApproximationDaK()) {
					intervallGl = addiere(intervallGl, g);
				} // Bei unterschiedlicher Approximation, Ganglinie ignorieren
			}

			verschiebe(intervallGl, tag.getTimeInMillis());
			auschneiden(intervallGl, intervall.getIntervall());
			teile.add(intervallGl);
		}

		// Ganglinien der Teilintervalle verketten
		for (final GanglinieMQ g : teile) {
			if (ganglinie == null) {
				ganglinie = g;
				ganglinie.setApproximationDaK(g.getApproximationDaK());
			} else {
				ganglinie = verbinde(ganglinie, g, Long.MAX_VALUE);
				if (ganglinie.getApproximationDaK() != g.getApproximationDaK()) {
					ganglinie.setApproximationDaK(0);
				}
			}
		}

		return ganglinie;
	}

	/**
	 * Bestimmt alle notwendigen Eingangsdaten der Prognose.
	 */
	private void bestimmeEingangsdaten() {
		prognoseParameter = prognoseParameterHelper.getPrognoseParameter();
		LogTools.log(log, new BetriebsmeldungDaten(anfrage.getMessQuerschnitt().getSystemObject()),
				GlProgMsg.FinerPrognoseParameter, anfrage.getMessQuerschnitt(), prognoseParameter);

		final AnalyseganglinieHelper analyseganglinieHelper = new AnalyseganglinieHelper();
		analyseganglinieHelper.setAnfrage(getAnfrage());
		analyseganglinieHelper.setPrognoseParameter(prognoseParameter);
		analyseganglinie = analyseganglinieHelper.getAnalyseganglinie();

		historischeGanglinien = new HistorischeGanglinienHelper(anfrage.getMessQuerschnitt())
				.getHistorischeGanglinien();
		LogTools.log(log, new BetriebsmeldungDaten(anfrage.getMessQuerschnitt().getSystemObject()),
				GlProgMsg.FinestHistotorischeGanglinien, anfrage.getMessQuerschnitt(),
				historischeGanglinien);

		final EreignisHelper ereignisHelper = new EreignisHelper();
		ereignisHelper.setAnfrage(getAnfrage());
		ereignisHelper.setAbsenderZeichen(absenderZeichen);
		kalenderantwort = ereignisHelper.getEreignisse();
	}

	/**
	 * Prüft, ob dies die Bearbeitung einer zyklischen Prognoseanfrage ist und ob sie noch gültig
	 * ist. Eine zyklische Prognose ist bis zum Ende des angefragten Prognosezeitraums gültig.
	 *
	 * @return {@code true}, wenn dies eine zyklische Prognose ist und der Prognosezeitraum noch
	 *         nicht abgelaufen ist.
	 */
	public boolean isZyklischAktiv() {
		return anfrage.isZyklischePrognose() && (anfrage.getPrognoseZeitraum()
				.getEnd() > ObjektFactory.getInstanz().getVerbindung().getTime());
	}

	@Override
	public String toString() {
		String s;

		s = getClass() + "["; //$NON-NLS-1$
		s += "absenderZeichen=" + absenderZeichen; //$NON-NLS-1$
		s += ", naechstePruefung=" + Timestamp.absoluteTime(naechstePruefung); //$NON-NLS-1$
		s += ", naechsteAktualisierung=" //$NON-NLS-1$
				+ Timestamp.absoluteTime(naechsteAktualisierung);
		s += ", anfrage=" + anfrage; //$NON-NLS-1$
		s += "]"; //$NON-NLS-1$

		return s;
	}

	/**
	 * Gibt den Wert des Flags für die Änderung des Prognoseparameters am Messquerschnitt zurück.
	 *
	 * @return {@code true}, wenn sich der Parameter geändert hat.
	 */
	public boolean isPrognoseParameterChanged() {
		return prognoseParameterChanged;
	}

	/**
	 * Setzt das Flag für die Änderung des Prognoseparameters am Messquerschnitt.
	 *
	 * @param prognoseParameterChanged
	 *            {@code true}, wenn sich der Parameter geändert hat.
	 */
	private void setPrognoseParameterChanged(final boolean prognoseParameterChanged) {
		this.prognoseParameterChanged = prognoseParameterChanged;
	}

}
