/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.glprog.glprog;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import com.bitctrl.util.Interval;

import de.bsvrz.sys.funclib.bitctrl.modell.kalender.objekte.EreignisTyp;
import de.bsvrz.sys.funclib.bitctrl.modell.kalender.parameter.PdEreignisTypParameter;

/**
 * Beschreibt ein Intervall im Prognosezeitraum. Ein Prognoseintervall wird durch seinen Zeitraum
 * und die im Zeitraum relevenaten Ganglinien beschrieben. Relevant sind alle Ganglinien zu im
 * Prognoseintervall anliegenden Ereignissen.
 *
 * @author BitCtrl Systems GmbH, Schumann
 */
public class PrognoseIntervall extends ArrayList<EreignisTypGanglinien> {

	/** Die ID für die Serialisierung. */
	private static final long serialVersionUID = 1L;

	/** Der Zeitraum des Prognoseintervalls. */
	private final Interval intervall;

	/**
	 * Konstruiert das Prognoseintervall.
	 *
	 * @param intervall
	 *            der Zeitraum des Prognoseintervalls.
	 */
	public PrognoseIntervall(final Interval intervall) {
		this.intervall = intervall;
	}

	/**
	 * Gibt die im Intervall gültigen Ereignistypen zurück, zu denen absolute Ganglinien existieren.
	 * Die Liste ist <em>absteigend</em> nach Priorität sortiert.
	 *
	 * @return die sortierte Liste von Tagesereignissen.
	 */
	public List<EreignisTyp> getEreignisTypenAbsolut() {
		final List<EreignisTyp> liste;

		// Alle Tagesereignisse sammeln ...
		liste = new ArrayList<>();
		for (final EreignisTypGanglinien eg : this) {
			if (eg.isTagesganglinie()) {
				liste.add(eg.getEreignisTyp());
			}
		}

		// ... und nach Priorität sortieren
		Collections.sort(liste, new Comparator<EreignisTyp>() {

			@Override
			public int compare(final EreignisTyp et1, final EreignisTyp et2) {
				PdEreignisTypParameter.Daten param1, param2;

				param1 = et1.getParameterDatensatz(PdEreignisTypParameter.class).abrufenDatum();
				param2 = et2.getParameterDatensatz(PdEreignisTypParameter.class).abrufenDatum();

				if (param1.getPrioritaet() > param2.getPrioritaet()) {
					return -1;
				} else if (param1.getPrioritaet() < param2.getPrioritaet()) {
					return 1;
				}
				return 0;
			}

		});

		return liste;
	}

	/**
	 * Gibt die Liste der Ereignistypen zurück, zu denen nur relative Ganglinien existieren.
	 *
	 * @return die Liste der Ereignistypen mit relativen Ganglinien.
	 */
	public List<EreignisTypGanglinien> getEreignisTypGanglinienRelativ() {
		List<EreignisTypGanglinien> liste;

		liste = new ArrayList<>();
		for (final EreignisTypGanglinien eg : this) {
			if (!eg.isTagesganglinie()) {
				liste.add(eg);
			}
		}

		return liste;
	}

	/**
	 * Gibt den Zeitraum des Prognoseintervall zurück.
	 *
	 * @return das Prognoseintervall.
	 */
	public Interval getIntervall() {
		return intervall;
	}

	/**
	 * Gibt die Tagesganglinie im Prognoseintervall zurück.
	 *
	 * @return die Tagesganglinie oder {@code null}, wenn keine existiert.
	 */
	public EreignisTypGanglinien getTagesEreignisTypGanglinien() {
		int zaehler;
		EreignisTypGanglinien ganglinie;

		zaehler = 0;
		ganglinie = null;
		for (final EreignisTypGanglinien eg : this) {
			if (eg.isTagesganglinie()) {
				ganglinie = eg;
				zaehler++;
			}
		}

		assert(zaehler == 0) || (zaehler == 1) : "Es darf maximal eine Tagesganglinie geben"; //$NON-NLS-1$
		return ganglinie;
	}

	@Override
	public String toString() {
		String s;

		s = getClass().getName() + "["; //$NON-NLS-1$
		s += "intervall=" + intervall; //$NON-NLS-1$
		s += ", ereignisTypGanglinien=" + super.toString(); //$NON-NLS-1$

		return s + "]"; //$NON-NLS-1$
	}

}
