/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.glprog.glprog;

import static com.bitctrl.Constants.MILLIS_PER_MINUTE;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

import java.sql.SQLException;
import java.util.Calendar;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;

import com.bitctrl.Constants;
import com.bitctrl.resource.Configuration;
import com.bitctrl.util.Interval;

import de.bsvrz.iav.gllib.gllib.Stuetzstelle;
import de.bsvrz.iav.gllib.gllib.dav.GanglinieMQ;
import de.bsvrz.iav.gllib.gllib.dav.GanglinienMQOperationen;
import de.bsvrz.iav.gllib.gllib.dav.GlProgAnfrage;
import de.bsvrz.iav.gllib.gllib.dav.Messwerte;
import de.bsvrz.iav.gllib.gllib.junit.EreignisFactory;
import de.bsvrz.iav.gllib.gllib.junit.GanglinienFactory;
import de.bsvrz.iav.gllib.gllib.modell.parameter.PdGanglinienModellPrognose;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;

/**
 * Führt den Prüffall 4 <em>Ganglinienprognose</em> aus.
 * <p>
 * Es wird eine einmalige langfristige Prognose durchgeführt. Für den morgigen Tag "TestMontag"
 * existieren drei relative Ganglinien (Cubic-Spline) für "TestMesse". Eine der drei ist als
 * Referenzganglinie ausgezeichnet. Es liegen keine aktuellen Messwerte (und Archivdaten) zum
 * Abfragezeitpunkt vor.
 * <p>
 * Die Prognoseganglinie wird mit drei anderen Ganglinien konkateniert:
 * <ol>
 * <li>mit einer sich überschneidenden Ganglinie,</li>
 * <li>mit einer die Prognoseganglinie berührenden Ganglinie und</li>
 * <li>mit einer weiter auseinanderliegenden Ganglinie.</li>
 * </ol>
 * <p>
 * Es wird ein Punkt der Prognoseganglinie berechnet.
 * <p>
 * Für den Test wird ein Messquerschnitt benötigt. Das Archiv wird nicht verwendet.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
@SuppressWarnings("nls")
public class Prueffall4 extends AbstractTestfall {

	/** Der Logger der Klassse. */
	private static Logger log = Logger.getLogger(Prueffall4.class.getName());

	/** Die Factory für die Kalenderdaten des Tests. */
	private static EreignisFactory ereignisFactory;

	/** Die Factory für die Ganglinien des Tests. */
	private static GanglinienFactory ganglinienFactory;

	/** Der Messquerschnitt an dem der Test stattfindet. */
	private static MessQuerschnittAllgemein mq;

	/**
	 * Führt Vorbereitungen durch, die einmalig vor dem Testfall notwendig sind.
	 * <ul>
	 * <li>Der Messquerschnitt für den Test wird bestimmt (siehe Configuration.properties).
	 * <em>Bereits vorhandene Ganglinien an diesem
	 * Messquerschnitt werden durch den Test überschrieben!</em></li>
	 * <li>Die Testereignisse und -ereignistypen werden angelegt.</li>
	 * <li>Die für den Test am Messquerschnitt benötigten Ganglinien werden parametriert (@code
	 * atg.ganglinie).</li>
	 * <li>Prüft ob der Prognoseparameter am Messquerschnitt gesetzt ist und ob als Auswahlkriterium
	 * "Referenzganglinie" angegeben ist. Ist der Parameter bereits vorhanden, wird lediglich die
	 * Auswahlmethode auf Referenzganglinie geändert.</li>
	 * </ul>
	 *
	 * @throws Exception
	 *             bei einem Fehler.
	 */
	@BeforeClass
	public static void beforeClass() throws Exception {
		final Configuration konfig;
		final Level level;
		final PdGanglinienModellPrognose param;
		PdGanglinienModellPrognose.Daten datum;

		konfig = Configuration.getConfiguration();
		level = Level.parse(konfig.getString("debug.level"));
		log.setLevel(level);

		DavUtil.ensureConnected();
		final ObjektFactory factory = ObjektFactory.getInstanz();

		// Testmessquerschnitt bestimmen und parametrieren
		mq = (MessQuerschnittAllgemein) factory
				.getModellobjekt(konfig.getString("prueffall4.objekt.mq"));
		assertNotNull("Der Testmessquerschnitt existiert nicht.", mq);

		// Ereignisse und Ereignistypen anlegen
		ereignisFactory = new EreignisFactory(
				"jdbc:hsqldb:res:de/bsvrz/iav/glprog/glprog/prueffall4;shutdown=true", "sa", "");
		if (konfig.getBoolean("reset")) {
			ereignisFactory.aufraeumen();
		}
		ereignisFactory.anlegenEreignistypen();
		ereignisFactory.anlegenEreignisse();

		// Ganglinien anlegen
		ganglinienFactory = new GanglinienFactory(mq,
				"jdbc:hsqldb:res:de/bsvrz/iav/glprog/glprog/prueffall4;shutdown=true", "sa", "",
				konfig.getString("dav.db"));
		if (konfig.getBoolean("reset")) {
			ganglinienFactory.aufraeumen();
			factory.getVerbindung().sleep(Constants.MILLIS_PER_SECOND);
		}
		ganglinienFactory.anlegenGanglinien(GanglinienFactory.ERSTER_TAG);

		// Prognoseparameter setzen bzw. aktualisieren
		param = mq.getParameterDatensatz(PdGanglinienModellPrognose.class);
		param.anmeldenSender();
		datum = param.abrufenDatum();
		if (datum.isValid()) {
			if (datum.getAuswahlMethode() != PdGanglinienModellPrognose.Daten.REFERENZGANGLINIE) {
				log.warning("Der Parameter existiert, muss aber angepasst werden.");
				datum.setAuswahlMethode(PdGanglinienModellPrognose.Daten.REFERENZGANGLINIE);
			}
		} else {
			datum = param.erzeugeDatum();
			datum.setAuswahlMethode(PdGanglinienModellPrognose.Daten.REFERENZGANGLINIE);
		}
		param.sendeDaten(datum);

		log.info("Warte 60 Sekunden ...");
		factory.getVerbindung().sleep(60 * Constants.MILLIS_PER_SECOND);
	}

	/**
	 * Räumt nach dem Test auf.
	 * <ul>
	 * <li>Die Testereignisse und -ereignistypen werden wieder gelöscht.</li>
	 * <li>Die Testganglinien am Messquerschnitt werden wieder entfernt.</li>
	 * </ul>
	 *
	 * @throws Exception
	 *             bei einem Fehler.
	 */
	@AfterClass
	public static void afterClass() throws Exception {
		ereignisFactory.aufraeumen();
		ganglinienFactory.aufraeumen();
	}

	/**
	 * Ruft nur den Superkonstruktor auf. Ist nur notwendig wegen der Exception die der
	 * Superkonstruktor werfen kann.
	 *
	 * @throws SQLException
	 *             bei einem Datenbankfehler.
	 */
	public Prueffall4() throws SQLException {
		super();
	}

	/**
	 * Führt eine einmalige Anfrage für eine langfristige Prognose an die Ganglinienprognose aus. Im
	 * Ergebnis werden nur die Verfahren der langfristigen Prognose verwendet. Die als
	 * Referenzganglinie ausgezeichnten Ganglinien werden für die Prognose verwendet.
	 *
	 * @throws Exception
	 *             bei einem Fehler.
	 */
	@Test(timeout = 60 * 1000)
	public void testPrognoseanfrage() throws Exception {
		ausfuehren();
		auswerten();
	}

	/**
	 * Testet die Konkatenation zweier sich schneidender Ganglinien. Die Konkatenation muss
	 * fehlschlagen.
	 */
	@Test(expected = java.lang.IllegalArgumentException.class)
	public void testKonkatenationSchneiden() {
		final GanglinieMQ ganglinie, g;

		log.info("Versuche zwei sich schneidende Ganglinien zu verbinden ...");

		ganglinie = getSollWerte().get(0).clone();
		ganglinie.setApproximationDaK(GanglinieMQ.APPROX_POLYLINE);

		g = new GanglinieMQ();
		g.setMessQuerschnitt(ganglinie.getMessQuerschnitt());
		for (final Stuetzstelle<Messwerte> s : ganglinie.getStuetzstellen()) {
			g.setStuetzstelle(s);
		}
		g.setApproximationDaK(GanglinieMQ.APPROX_POLYLINE);

		// Um 10 Minuten ineinander verschieben
		GanglinienMQOperationen.verschiebe(g, -10 * MILLIS_PER_MINUTE);
		GanglinienMQOperationen.verbinde(ganglinie, g, Long.MAX_VALUE);
	}

	/**
	 * Testet die Konkatenation zweier sich berührender Ganglinien. Im Ergebnis werden beide
	 * Ganglinien verbunden. Der Berührungspunkt ist der Mittelwert der beiden y-Werte.
	 */
	@Test
	public void testKonkatenationBeruehren() {
		final GanglinieMQ ganglinie, g, result;
		final Stuetzstelle<Messwerte> s1, s2, soll, ist;
		final Double qKfz, qLkw, vPkw, vLkw;

		log.info("Versuche zwei sich berührende Ganglinien zu verbinden ...");

		ganglinie = getSollWerte().get(0).clone();
		ganglinie.setApproximationDaK(GanglinieMQ.APPROX_POLYLINE);

		g = new GanglinieMQ();
		g.setMessQuerschnitt(ganglinie.getMessQuerschnitt());
		for (final Stuetzstelle<Messwerte> s : ganglinie.getStuetzstellen()) {
			g.setStuetzstelle(s);
		}

		// Um 24 Stunden versetzt verschieben
		GanglinienMQOperationen.verschiebe(g, 24 * Constants.MILLIS_PER_HOUR);
		try {
			result = GanglinienMQOperationen.verbinde(ganglinie, g, Long.MAX_VALUE);
		} catch (final IllegalArgumentException ex) {
			log.severe(
					"Der Versuch zwei sich berührende Ganglinien zu verbinden schlug fehl. " + ex);
			System.exit(-1);
			return;
		}
		result.setApproximationDaK(GanglinieMQ.APPROX_POLYLINE);

		// Stützstellen prüfen
		s1 = g.getStuetzstellen().get(0);
		s2 = ganglinie.getStuetzstelle(s1.getZeitstempel());

		qKfz = (s1.getWert().getQKfz() + s2.getWert().getQKfz()) / 2;
		qLkw = (s1.getWert().getQLkw() + s2.getWert().getQLkw()) / 2;
		vPkw = (s1.getWert().getVPkw() + s2.getWert().getVPkw()) / 2;
		vLkw = (s1.getWert().getVLkw() + s2.getWert().getVLkw()) / 2;
		soll = new Stuetzstelle<>(g.getStuetzstellen().get(0).getZeitstempel(),
				new Messwerte(qKfz, qLkw, vPkw, vLkw));

		ist = result.getStuetzstelle(g.getStuetzstellen().get(0).getZeitstempel());
		assertEquals(
				"Die Stützstelle im Berührungspunkt muss dem Durchschnitt der Werte entsprechen.",
				soll, ist);
	}

	/**
	 * Testet die Konkatenation zweier weiter entfernt liegender Ganglinien. Im Ergebnis ist der
	 * Bereich zwischen den beiden Ganglinien undefiniert.
	 */
	@Test
	public void testKonkatenationEntfernt() {
		final GanglinieMQ ganglinie, g, result;

		log.info("Versuche zwei entfernte Ganglinien zu verbinden ...");

		ganglinie = getSollWerte().get(0).clone();
		ganglinie.setApproximationDaK(GanglinieMQ.APPROX_POLYLINE);

		g = new GanglinieMQ();
		g.setMessQuerschnitt(ganglinie.getMessQuerschnitt());
		for (final Stuetzstelle<Messwerte> s : ganglinie.getStuetzstellen()) {
			g.setStuetzstelle(s);
		}

		// Um 48 Stunden versetzt verschieben
		GanglinienMQOperationen.verschiebe(g, 48 * Constants.MILLIS_PER_HOUR);
		try {
			result = GanglinienMQOperationen.verbinde(ganglinie, g, Constants.MILLIS_PER_HOUR);
		} catch (final IllegalArgumentException ex) {
			log.severe("Der Versuch zwei entfernte Ganglinien zu verbinden schlug fehl. " + ex);
			System.exit(-1);
			return;
		}
		result.setApproximationDaK(GanglinieMQ.APPROX_POLYLINE);

		// Prüfe den Bereich zwischen den Ganglinien
		assertEquals("Die Stützstelle zwischen den Ganglinien muss undefiniert (null) sein.", null,
				result.getStuetzstelle(ganglinie.lastKey() + 1).getWert().getQKfz());
		assertEquals("Die Stützstelle zwischen den Ganglinien muss undefiniert (null) sein.", null,
				result.getStuetzstelle(ganglinie.lastKey() + 1).getWert().getQLkw());
		assertEquals("Die Stützstelle zwischen den Ganglinien muss undefiniert (null) sein.", null,
				result.getStuetzstelle(ganglinie.lastKey() + 1).getWert().getVPkw());
		assertEquals("Die Stützstelle zwischen den Ganglinien muss undefiniert (null) sein.", null,
				result.getStuetzstelle(ganglinie.lastKey() + 1).getWert().getVLkw());
	}

	/**
	 * Berechnet einen Punkt der Ganglinie. Der Punkt muss mit dem vorher berechneten Werten
	 * übereinstimmen.
	 */
	@Test
	public void testPunktDerGanglinieAusrechnen() {
		final GanglinieMQ ganglinie;
		final Stuetzstelle<Messwerte> soll, ist;

		log.info("Berechne einen Punkt der Ganglinie ...");

		soll = new Stuetzstelle<>(8 * Constants.MILLIS_PER_HOUR,
				new Messwerte(105.0, 10.0, 175.0, 55.0));

		ganglinie = getSollWerte().get(0);
		ist = ganglinie.getStuetzstelle(8 * Constants.MILLIS_PER_HOUR);

		assertEquals(soll, ist);
	}

	@Override
	protected List<GlProgAnfrage> abfragenAnfragen() throws SQLException {
		GlProgAnfrage anfrage;
		long start, ende;
		Calendar calendar;

		calendar = Calendar.getInstance();
		calendar.set(Calendar.HOUR_OF_DAY, 24);
		calendar.set(Calendar.MINUTE, 0);
		calendar.set(Calendar.SECOND, 0);
		calendar.set(Calendar.MILLISECOND, 0);
		start = calendar.getTimeInMillis();
		calendar.add(Calendar.HOUR_OF_DAY, 24);
		ende = calendar.getTimeInMillis();

		anfrage = new GlProgAnfrage(mq, new Interval(start, ende), true);

		return Collections.singletonList(anfrage);
	}

	@Override
	protected List<GanglinieMQ> abfragenAntwortenSoll() throws SQLException {
		List<GanglinieMQ> soll;
		Iterator<GanglinieMQ> iterator;

		soll = ganglinienFactory.getGanglinienTestdaten(2);
		iterator = soll.iterator();
		while (iterator.hasNext()) {
			if (iterator.next().size() == 0) {
				iterator.remove();
			}
		}

		assertEquals("Die Anzahl der Prognoseganglinie ist falsch.", 1, soll.size());
		return soll;
	}

}
